module NestedStruct
  class Coercer
    attr_reader :type, :target

    def initialize(expression)
      if expression.is_a?(Class)
        @type = :single
        @target = expression
      elsif expression.is_a?(Array)
        raise Errors::InvalidCoercerExpression if expression.size != 1
        @type = :multiple
        @target = expression.first
      else
        raise Errors::InvalidCoercerExpression
      end

      raise Errors::InvalidCoercerTarget if !valid_target?
    end

    def coerce(value)
      if multiple?
        raise Errors::InvalidCoercerValue if !value.is_a?(Array)
        value.map { |e| single_coercion(e) }
      else
        single_coercion(value)
      end
    end

    def multiple?
      type == :multiple
    end

    private

    def valid_target?
      target.ancestors.include?(NestedStruct::Interface)
    end

    def single_coercion(hash)
      raise Errors::InvalidCoercerValue if !hash.is_a?(Hash)
      target.new(hash)
    end
  end
end
