module NestedStruct
  module Interface
    def self.included(base)
      base.extend(ClassMethods)
    end

    module ClassMethods
      attr_reader :field_names, :fields

      def field(name, coercer_expression = nil)
        if superclass.ancestors.include?(NestedStruct::Interface)
          @field_names ||= superclass.field_names.dup || []
          @fields ||= superclass.fields.dup || []
        else
          @field_names ||= []
          @fields ||= []
        end

        if !field_names.include?(name)
          @field_names << name
          @fields << Field.new(name, coercer_expression)
          attr_accessor(name)
        end
      end
    end

    def initialize(attributes = {})
      assign_field_values(attributes)
    end

    def to_hash
      Hash[map_fields]
    end

    private

    def map_fields
      mapped_fields = []

      self.class.fields.each do |field|
        result = [field.name.to_sym, primitivize(public_send(field.name))]
        mapped_fields << result if !result.last.is_a?(NestedStruct::Empty)
      end

      mapped_fields
    end

    def assign_field_values(attributes = {})
      self.class.fields.each do |field|
        instance_variable_set(
          "@#{field.name}",
          field_value(field, attributes)
        )
      end
    end

    def primitivize(value)
      return nil if value.nil?
      return value.map(&:to_hash) if value.is_a?(Array)
      return value.to_hash if value.respond_to?(:to_hash)
      value
    end

    def field_value(field, attributes)
      if NestedStruct.configuration.use_empty_value
        field.with(attributes.fetch(field.name, NestedStruct::EMPTY))
      else
        field.with(attributes[field.name])
      end
    end
  end
end
